import jinja2
from enum import Enum
import yaml
from pyspark.sql.types import StructType
from typing import Dict
import os
import pkg_resources
import logging


YETL_CONFIG = "YETL_CONFIG"
_ENCODING = "utf-8"
_DBX_WORKSPACE_PATH = "/Workspace"
_DBX_REPO_PATH = "/Workspace/Repos"


class VersionNotFoundException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class JinjaVariables(Enum):
    DATABASE = "database"
    TABLE = "table"
    CHECKPOINT = "checkpoint"
    FILENAME_DATE_FORMAT = "filename_date_format"
    PATH_DATE_FORMAT = "path_date_format"
    CONTAINER = "container"


def is_databricks():
    return "DATABRICKS_RUNTIME_VERSION" in os.environ


def check_version(data: dict):
    _logger = logging.getLogger(__name__)
    version = data.get("version")

    if version is None:
        raise VersionNotFoundException()

    del data["version"]

    pkg_version = pkg_resources.get_distribution("dbxconfig").version
    pkg_version = pkg_version.split(".")

    try:
        version = version.split(".")
        if pkg_version[0] != version[0] or pkg_version[1] != version[1]:
            _logger.warning(
                f"Configuration and library shows that it's incompatible config version config_version={version} package_version={pkg_version}"
            )

    except Exception as e:
        raise VersionNotFoundException from e


def render_jinja(data: str, replacements: Dict[JinjaVariables, str]):
    _logger = logging.getLogger(__name__)
    _logger.debug(f"Rendering Jinja string {data}")
    if data and isinstance(data, str):
        replace = {k.value: v for (k, v) in replacements.items()}
        skip = False
        for k, v in replace.items():
            if v is None and "{{" + k + "}}" in data.replace(" ", ""):
                skip = True
                break

        if not skip:
            template: jinja2.Template = jinja2.Template(data)
            data = template.render(replace)
    _logger.debug(f"Rendered Jinja string {data}")

    return data


def abs_config_path(root: str, path: str):
    _logger = logging.getLogger(__name__)
    if not os.path.isabs(path):
        path = os.path.join(root, path)
    _logger.debug(f"Absolute config path {path}")
    return path


def get_config_path(project: str, path: str):
    _logger = logging.getLogger(__name__)

    default_path = "."
    if is_databricks():
        if os.path.exists(_DBX_WORKSPACE_PATH):
            default_path = f"{_DBX_WORKSPACE_PATH}"

        if os.getcwd().startswith(_DBX_REPO_PATH):
            repo_path = "/".join(os.getcwd().split("/")[0:5])
            default_path = repo_path

    if not path:
        path = os.getenv(YETL_CONFIG, default_path)
    path = os.path.abspath(path)
    path = os.path.join(path, project)
    _logger.info(f"Absolute root config path {path}")
    return path


def load_schema(path: str):
    _logger = logging.getLogger(__name__)
    schema = load_yaml(path)
    _logger.info(f"Loading schema {path}")
    schema = StructType.fromJson(schema)

    return schema


def load_yaml(path: str):
    _logger = logging.getLogger(__name__)
    _logger.info(f"Loading yaml file {path}")
    with open(path, "r", encoding=_ENCODING) as f:
        try:
            data = yaml.safe_load(f)
        except yaml.YAMLError as e:
            location = ""
            if hasattr(e, "problem_mark"):
                mark = e.problem_mark
                location = f"Error position ({mark.line}, {mark.column})"

            if hasattr(e, "problem"):
                problem = f"{e.problem}."

            raise Exception(f"Invalid yaml format in {path}. {problem} {location}")
    _logger.debug(data)
    return data


def get_ddl(spark_schema: StructType, header: bool = True):
    _logger = logging.getLogger(__name__)
    _logger.debug(f"Converting spark schema to ddl with header={str(header)}")
    if header:
        ddl = [f"{f.name} {f.dataType.simpleString()}" for f in spark_schema.fields]
        _logger.debug(ddl)
    else:
        ddl = [
            f"_c{i} {f.dataType.simpleString()}"
            for i, f in enumerate(spark_schema.fields)
        ]
        _logger.debug(ddl)

    return ddl


def get_html_table(data: dict):
    rows = []
    for k, v in data.items():
        if isinstance(v, dict):
            for ki, vi in v.items():
                row = f"<tr><td>{k}.{ki}</td><td>{vi}</td></tr>"
                rows.append(row)
        else:
            row = f"<tr><td>{k}</td><td>{v}</td></tr>"
            rows.append(row)

    html = "".join(rows)

    html = f"""
  <table>
    <tr><th>Name</th><th>Source</th><th>Destination</th></tr>
    {html}
  </table>
  """

    return html



# implicit, not referenced - must be the 1st import
from ._logging_config import configure_logging
import logging
from ._config import Config
from ._timeslice import Timeslice
from ._stage_type import StageType


def yetl_flow(
    stage: StageType,
    project: str,
    pipeline: str = None,
    config_path: str = None,
):
    def decorate(function):
        def wrap_function(*args, **kwargs):
            configure_logging(project)
            _logger = logging.getLogger(__name__)

            _pipeline = pipeline
            if not _pipeline:
                _pipeline = function.__name__

            _logger.info(f"Loading pipeline configuration {_pipeline}")

            timeslice = kwargs.get("timeslice", Timeslice(day="*", month="*", year="*"))
            if "timeslice" in kwargs.keys():
                del kwargs["timeslice"]

            try:
                table = kwargs["table"]
                del kwargs["table"]
            except KeyError as e:
                raise Exception(f"{e} is a required argument for a yetl flow function")

            config = Config(
                project=project, pipeline=_pipeline, config_path=config_path
            )
            table_mapping = config.get_table_mapping(
                timeslice=timeslice, stage=stage, table=table
            )
            _logger.info(f"Calling function {function.__name__}")
            ret = function(
                *args,
                table_mapping=table_mapping,
                **kwargs,
            )
            return ret

        return wrap_function

    return decorate



from concurrent.futures import ThreadPoolExecutor, Future, as_completed
from ._notebook import Notebook
from typing import List


# execute a notebook using databricks workflows
def _execute_notebook(notebook: Notebook, dbutils):
    """Execute a notebookd using databricks workflows"""
    msg = {
        "_message": f"Executing notebook {notebook.path} parameters {notebook.parameters}",
        "status": "executing",
        "notebook": notebook.path,
    }
    print(msg["_message"], flush=True)

    try:
        result = dbutils.notebook.run(
            notebook.path, notebook.timeout, notebook.parameters
        )
        msg = {
            "_message": f"Succeeded notebook {notebook.path}",
            "status": "succeeded",
            "notebook": notebook.path,
        }
        print(msg["_message"], flush=True)
        return result

    except Exception as e:
        if notebook.retry < 1:
            msg = {
                "_message": f"notebook {notebook.path} failed.",
                "status": "failed",
                "error": str(e),
                "notebook": notebook.path,
            }
            print(msg["_message"], flush=True)
            raise Exception(msg["_message"])

        msg = {
            "_message": f"Retrying notebook {notebook.path}",
            "status": "executing",
            "notebook": notebook.path,
        }
        print(msg["_message"], flush=True)
        notebook.retry -= 1


def _try_future(future: Future, catch=False):
    return future.result()


def execute_notebooks(notebooks: List[Notebook], maxParallel: int, dbutils):
    msg = {
        "_message": f"Executing {len(notebooks)} with maxParallel of {maxParallel}",
        "notebooks": len(notebooks),
        "maxParallel": maxParallel,
    }
    print(msg["_message"], flush=True)

    with ThreadPoolExecutor(max_workers=maxParallel) as executor:
        results = [
            executor.submit(_execute_notebook, notebook, dbutils)
            for notebook in notebooks
            if notebook.enabled
        ]

        # the individual notebooks handle their errors and pass back a packaged result
        # we will still need to handle the fact that the notebook execution call may fail
        # or a programmer missed the handling of an error in the notebook task
        # that's what tryFuture(future:Future) does
        results_list = [_try_future(r) for r in as_completed(results)]

        print(
            f"Finished executing {len(notebooks)} with maxParallel of {maxParallel}",
            flush=True,
        )
        return results_list



from pydantic import BaseModel, Field
from typing import Union, Any, Dict
from .dataset import Table


class TableMapping(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    destination: Table = Field(...)
    source: Union[Dict[str, Table], Table] = Field(...)



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="dbxconfig",
    version="5.0.6",
    description="Databricks Configuration Framework",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://dbxconfig.readthedocs.io/en/latest/",
    project_urls={
        'GitHub': 'https://github.com/semanticinsight/dbxconfig',
        'Documentation': 'https://dbxconfig.readthedocs.io/en/latest/'
    },
    author="Shaun Ryan",
    author_email="shaun_chiburi@hotmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=[
        "dbxconfig",
        "dbxconfig.dataset",
        "dbxconfig.workflow"
    ],
    install_requires=[
          'PyYAML',
          'jinja2',
          'pydantic'
      ],
    zip_safe=False
)



# used to carry notebook data
from pydantic import BaseModel, Field
from typing import Any


class Notebook(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        # add the notebook path to parameters for error reporting.
        self.parameters["notebook"] = self.path

    path: str = Field(...)
    timeout: int = Field(default=3600)
    parameters: dict = Field(default={})
    retry: int = Field(default=0)
    enabled: bool = Field(default=True)
